/***********************************************************************
sample.c --

This is a sample application program for DOS which demonstrates
how to make use of the rckt.h and rckt.c low-level code.  This program
is built under BORLAND C v3.1 and really does no useful work by itself.
It opens all the ports on a RocketPort board and starts them transmitting
data.  It then also checks each port for receive data and if any
exists, pulls it in and displays the count of how many bytes were
received.  You might hook up loopback plugs on each port so that
what gets transmitted, is also received and counted.  The intention
of the program is to provide sample code that calls into the toolkit
routines and also to give you a first program to port to your platform
that will show some functionality of the board and toolkit once the
port is complete.

This particular application sets the RocketPort board up to issue a
periodic interrupt (137 Hz) which then causes receive and transmit
work routines to be called depending on what work needs to be done
during that individual tick interval.  This model is similar to
how a Unix driver might work, where the periodic interrupt is actually
a polling mechanism provided by the OS kernel. In this case, the
board would not need to issue a ticker interrupt at all.

The application also illustrates the use of a double buffer mechanism.
Here, host memory is used to provide bigger queues than are available
on the board itself.  The board provides a 1K receive and 255 byte
transmit FIFO for each port.  Sometimes these sizes are too small
for some applications.  This code then shows how to move data to and
from the host queue during the polling time and how to deal with
receive flow control when these secondary buffers are in use.

Copyright 1994-1996 Comtrol Corporation.   All Rights Reserved.
***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <conio.h>

#define TRUE 1
#define FALSE 0

#include "rcktldef.h"
#include "rckt.h"

#define ESUCCESS	0
#define EFAILURE	-1

#define MDMSTA_DCD	1
#define MDMSTA_DSR	2
#define MDMSTA_CTS	4
#define MDMCTL_DTR	1
#define MDMCTL_RTS	2

#define ASCII_XON 0x11
#define ASCII_XOFF 0x13

//	RCKT Port Data Structure

typedef struct
{
	int	Available;
	int	Acquired;

	CHANNEL_T	*cp;

	unsigned char	*txq;			// pointer to start of txq
	int	txqPut;
	int	txqGet;
	int	txqSize;		// size of the transmit queue
	int	txqCount;		// current count of chars in txq

	unsigned char	*rxq;			// pointer to start of rxq
	int	rxqPut;
	int	rxqGet;
	int	rxqSize;		// size of the receive queue
	int	rxqCount;		// current count of chars in rxq
	int	rxqFlowOff;		// receive flow off point (80% typ)
	int	rxqFlowOn;		// receive flow on point (20% typ)
	unsigned char	rxFlowState;		// flow on/off state of rx

	unsigned char	transmitXON;		// what starts our transmitter
	unsigned char	transmitXOFF;		// what stops our transmitter
	unsigned char	receiveXON;		// what we send when we can receive
	unsigned char	receiveXOFF;		// what we send when we can't receive

	long bitRate;		// bit rate index

	unsigned char	dataBits;		// bits/char
	unsigned char	stopBits;		// number of stop bits
	unsigned char	parityMode;		// parity mode index
	unsigned char	FlowCtrlMode;

	unsigned long	framingErrors;		// accumulated errors
	unsigned long	parityErrors;
	unsigned long	overrunSoftware;
	unsigned long	overrunHardware;
} RCKTPort_t;

typedef struct {		// a struct to hold IRQ config
	BYTE Mask;		// information
	BYTE IntNum;
} IRQData_t;

IRQData_t IRQData[16]= {
	{0,0},		/* IRQ0 */
	{0,0},		/* IRQ1 */
	{0,0},		/* IRQ2 */
	{0xf7,0xb},	/* IRQ3 */
	{0xef,0xc},	/* IRQ4 */
	{0xdf,0xd},	/* IRQ5 */
	{0,0},		/* IRQ6 */
	{0,0},		/* IRQ7 */
	{0,0},		/* IRQ8 */
	{0xfd,0x71},	/* IRQ9 */
	{0xfb,0x72},	/* IRQ10 */
	{0xf7,0x73},	/* IRQ11 */
	{0xef,0x74},	/* IRQ12 */
	{0,0},	 	/* IRQ13 */
	{0,0},		/* IRQ14 */
	{0x7f,0x77}	/* IRQ15 */
};

RCKTPort_t		RcktPort[MAX_PORTS_PER_BOARD];
unsigned int		IOAddr, IrqNum, MaxPorts, nPorts;
int			OpenPortCount, OpenPortList[MAX_PORTS_PER_BOARD];

int RCKTInit(void);
void RCKTReset(void);
int RCKTConfigure(int);
int RCKTOpen(int);
int RCKTClose(int);
int RCKTModemControl(int, int, int);
int RCKTModemStatus(int);
int RCKTReadData(int, char *, int, int *);
int RCKTWriteData(int, char *, int, int *);
int RCKTTxFlush(int PortNo);
int RCKTRxFlush(int PortNo);
void interrupt RCKTPollISR();

unsigned htoi(char *);
void ExitProgram(int);

/*-------------------------------------------------------------------------
 main - 
|-------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
	int PICio;
	unsigned char mask;
	int pn, i;
	RCKTPort_t *Port;
	char *Msg="The quick brown fox jumps over the lazy dogs back\n";
	int NumWritten, NumRead, ToRead;
	char buf[128];

	if (argc < 3)
	{	fprintf(stderr,"Usage: sample IOAddr IrqNum [maxports]\n");
		exit(-1);
	}

	IOAddr= htoi(argv[1]);
	IrqNum= atoi(argv[2]);

	if (argc == 4)
		MaxPorts= atoi(argv[3]);
	else
		MaxPorts= MAX_PORTS_PER_BOARD;

//// Setup the vector and PIC for the choosen IRQ and enable interrupts ////

	_disable();
		setvect(IRQData[IrqNum].IntNum, RCKTPollISR);
		PICio= (IrqNum > 7) ? 0xa1 : 0x21;
		mask= inp(PICio);
		mask &= IRQData[IrqNum].Mask;
		outp(PICio, mask);
	_enable();

//// Initialize the Rocket board at the IOAddr and IrqNum specified

	printf("IOAddr= %x, IrqNum= %d\n",IOAddr, IrqNum);

	if (RCKTInit() != ESUCCESS)
	{
		fprintf(stderr,"RCKTInit() failed.\n");
		ExitProgram(-1);
	}

	if (MaxPorts < nPorts)
		nPorts= MaxPorts;
		
	printf("RocketPort board initialized. io= 0x%03x, irq= %d, nPorts= %d\n",
			IOAddr, IrqNum, nPorts);

//// Initialize all the ports to 38400, N, 8, 1 with 1K transmit and
//// receive buffers and software flow control

	for (pn=0; pn<nPorts; pn++)
	{
		Port= &RcktPort[pn];

		Port->rxqSize= 1024;
		Port->txqSize= 1024;

		printf("Setting up port %d...",pn);

		if (RCKTOpen(pn) != ESUCCESS)
		{	printf("RCKTOpen(%d) failed.\n",pn);
			ExitProgram(-1);
		}

		Port->bitRate= 38400L;
		Port->dataBits= 8;
		Port->stopBits= 1;
		Port->parityMode= 0;
		Port->FlowCtrlMode= 2;		// sw flow control

		if (RCKTConfigure(pn) != ESUCCESS)
		{	printf("RCKTConfigure(%d) failed.\n",pn);
			ExitProgram(-1);
		}

		//// assert DTR and RTS on each port too

		if (RCKTModemControl(pn, MDMCTL_DTR|MDMCTL_RTS, 1) != ESUCCESS)
		{	printf("RCKTModemControl(%d,...) failed.\n",pn);
			ExitProgram(-1);
		}

		printf("done.\n");
	}

//// some really meaningless work to do but it demostrates transmit
//// and receive data functions

	while (1)
	{
		// write a message to all the ports

		for (pn=0; pn<nPorts; pn++)
			RCKTWriteData(pn,Msg,strlen(Msg),&NumWritten);

		// check which ports have receive data and pull it in
		// one buf at a time

		for (pn=0; pn<nPorts; pn++)
		{	if (RcktPort[pn].rxqCount)
			{	ToRead= RcktPort[pn].rxqCount;
				RCKTReadData(pn,buf,sizeof(buf),&NumRead);
				printf("pn= %d, ToRead= %d, NumRead= %d\n",
					pn,ToRead,NumRead);
			}
		}

		if (kbhit())
			break;
	}

	ExitProgram(0);
	return(0);		// will never get here.. make compiler happy
}

/*-------------------------------------------------------------------------
 ExitProgram - 
|-------------------------------------------------------------------------*/
void ExitProgram(int ErrorStatus)
{
	unsigned int PICio;
	unsigned char mask;

	RCKTReset();		// take the board down to the ground
				// no more ticking or anything

//	_disable();		// disable the IRQ we were using too
//		PICio= (IrqNum > 7) ? 0xa1 : 0x21;
//		mask= inp(PICio);
//		mask |= IRQData[IrqNum].Mask;
//		outp(PICio, mask);
//	_enable();

	exit(ErrorStatus);
}

/*-------------------------------------------------------------------------
 RCKTInit - Initialize the RocketPort board at the IOAddr and IrqNum that are
   globally defined
|-------------------------------------------------------------------------*/
int RCKTInit()
{
	WORD mbio;
	int ctlid;
	int aiop, nch, pn;
	WORD aiopio[MAX_AIOPS_PER_BOARD];
	CONTROLLER_T *ctlp;
	
//// establish ioaddr for MUDBAC and all AIOPs

	mbio= IOAddr + 0x40;
	for (aiop=0; aiop<MAX_AIOPS_PER_BOARD; aiop++)
		aiopio[aiop]= IOAddr + (aiop * 0x400);

	ctlp= &sController[0];
	nch= 0;

//// init the board.  Enable it to generate 137 Hz tick but no other interrupt
//// types

	ctlid= sInitController(ctlp, 0, mbio, aiopio,
		AIOP_CTL_SIZE, IrqNum, FREQ_137HZ, TRUE);

//// if the board successfully came up, then reset each AIOP to ground zero
//// and enable it.  Once each AIOP is enabled, ask it how many channels
//// it has (most have eight; but there is a 4 port case).

	if (ctlid != CTLID_NULL)
	{	for (aiop=0; aiop<sGetControllerNumAiop(ctlp); aiop++)
		{
			sEnAiop(ctlp, aiop);
			sResetAiopByNum(ctlp, aiop);
           nch += ctlp->AiopNumChan[aiop];
		}
	}

	if (ctlid == CTLID_NULL || (nch != 8 && nch != 16 && nch != 32))
		return(EFAILURE);

	nPorts= nch;
	if (MaxPorts < nch)
		nch= MaxPorts;

//// mark which ports are available in our little struct that keeps
//// track of them

	for (pn=0; pn<MAX_PORTS_PER_BOARD; pn++)
	{	memset(&RcktPort[pn], 0, sizeof(RCKTPort_t));
		if (pn < nch)
			RcktPort[pn].Available= 1;
	}

	OpenPortCount= 0;

//// enable global interrupts from the board.  This is what turns on the
//// driver connected to the IRQ line on the bus.

	sEnGlobalInt(ctlp);

	return(ESUCCESS);
}

/*-------------------------------------------------------------------------
 RCKTReset - Reset the board to ground zero and make sure its interrupts are
  disabled
|-------------------------------------------------------------------------*/
void RCKTReset()
{
	int aiop;
	CONTROLLER_T *ctlp;

	ctlp= &sController[0];

	sDisGlobalInt(ctlp);

	for (aiop=0; aiop<sGetControllerNumAiop(ctlp); aiop++)
		sResetAiopByNum(ctlp, aiop);

	OpenPortCount= 0;
}

/*-------------------------------------------------------------------------
  RCKTOpen -
Open one port on the board.  This particular routine will malloc some
space for transmit and receive queues and for the channel structure.
It will also initialize the channel within the AIOP and fill in entries
in the structure to their default settings.
|-------------------------------------------------------------------------*/
int RCKTOpen(int PortNo)
{
	CONTROLLER_T *ctlp;
	CHANNEL_T *cp;
	RCKTPort_t *Port;
	int pn, aiop, ch;

	Port= &RcktPort[PortNo];

	if ((!Port->Available) || Port->Acquired)
		return(EFAILURE);

	Port->rxqPut= Port->rxqGet= Port->txqCount= 0;
	Port->rxq= (unsigned char *)malloc(Port->rxqSize);
	if (Port->rxq == NULL)
		return(EFAILURE);

	Port->txqPut= Port->txqGet= Port->txqCount= 0;
	Port->txq= (unsigned char *)malloc(Port->txqSize);
	if (Port->txq == NULL)
		return(EFAILURE);

	Port->cp= (CHANNEL_T *)malloc(sizeof(CHANNEL_T));
	if (Port->cp == NULL)
		return(EFAILURE);

	Port->Acquired= 1;

//// compute the flow off and flow on points in the rx queue

	Port->rxqFlowOff= (Port->rxqSize / 10) * 8;	// 80 %
	Port->rxqFlowOn= (Port->rxqSize / 10) * 2;	// 20 %
	Port->rxFlowState= 0;

	Port->framingErrors= Port->parityErrors=
	Port->overrunSoftware= Port->overrunHardware= 0L;

	cp= Port->cp;

	ctlp= &sController[0];
	aiop= PortNo / MAX_PORTS_PER_AIOP;	// which AIOP
	ch= PortNo % MAX_PORTS_PER_AIOP;	// which channel on that AIOP

	sSetDevMap(0, aiop, ch, PortNo);
	sInitChan(ctlp, cp, aiop, ch);

	// set the defaults for the port

	Port->bitRate= 9600L;
	Port->dataBits= 8;
	Port->stopBits= 1;
	Port->parityMode= 0;

	Port->transmitXON= ASCII_XON;
	Port->transmitXOFF= ASCII_XOFF;
	Port->receiveXON= ASCII_XON;
	Port->receiveXOFF= ASCII_XOFF;

	Port->FlowCtrlMode= 0;

	RCKTTxFlush(PortNo);
	RCKTRxFlush(PortNo);

	sSetRxTrigger(cp, TRIG_1);
	sGetChanStatus(cp);
	sDisRxStatusMode(cp);

	sEnInterrupts(cp, (MCINT_EN|RXINT_EN|SRCINT_EN|CHANINT_EN));

	sGetChanStatus(cp);
	sDisRxStatusMode(cp);
	sClrTxXOFF(cp);

	sDisCTSFlowCtl(cp);
	sDisTxSoftFlowCtl(cp);

	sEnRxFIFO(cp);
	sEnTransmit(cp);

	// build a compact, ordered list of open ports

	_disable();	// don't let the ISR do anything during this adventure
	OpenPortCount= 0;
	for (pn=0; pn<MAX_PORTS_PER_BOARD; pn++)
	{	if (RcktPort[pn].Acquired)
		{	OpenPortList[OpenPortCount]= pn;
			OpenPortCount++;
		}
	}
	_enable();

	return(ESUCCESS);
}

/*-------------------------------------------------------------------------
 RCKTConfigure - Configure a port to the baud, bits, stops, flow control
   mode set in port structure
|-------------------------------------------------------------------------*/
int RCKTConfigure(int PortNo)
{
	CHANNEL_T *cp;
	RCKTPort_t *Port;

	Port= &RcktPort[PortNo];

	if (!Port->Acquired)
		return(EFAILURE);

	cp= Port->cp;

		// baud rate

   sSetBaudRate(cp, Port->bitRate, 1);

		// bits per character: 7=7, all others= 8

	sSetData8(cp);
	if (Port->dataBits == 7)
		sSetData7(cp);

		// parity mode: 0=none, 1=odd, 2=even

	sDisParity(cp);
	if (Port->parityMode > 0)
	{	sEnParity(cp);
	
		if (Port->parityMode == 1)
			sSetOddParity(cp);

		if (Port->parityMode == 2)
			sSetEvenParity(cp);
	}

		// stop bits: 2=2, all others= 1

	sSetStop1(cp);
	if (Port->stopBits == 2)
		sSetStop2(cp);

	sSetTxXOFFChar(cp,Port->transmitXOFF);
	sSetTxXONChar(cp,Port->transmitXON);

	if (Port->FlowCtrlMode & 1)		// hardware flow enabled ?
	{
		if (Port->rxqCount >= Port->rxqFlowOff)
		{
			sClrRTS(cp);
			Port->rxFlowState |= 2;
		}
		sEnCTSFlowCtl(cp);
	}
	else
	{
		if (Port->rxFlowState & 2)
		{
			sSetRTS(cp);
			Port->rxFlowState &= ~2;
		}
		sDisCTSFlowCtl(cp);
	}

	if (Port->FlowCtrlMode & 2)		// software flow enabled ?
	{
		if (Port->rxqCount > Port->rxqFlowOff)
		{	sWriteTxPrioByte(cp, Port->receiveXOFF);
			Port->rxFlowState |= 1;
		}

		// enable output SW flow control

		sEnTxSoftFlowCtl(cp);
	}
	else
	{
		if (Port->rxFlowState & 1)
		{	sWriteTxPrioByte(cp, Port->receiveXON);
			Port->rxFlowState &= ~1;
		}

		// disable output SW flow control

		sDisTxSoftFlowCtl(cp);
		sClrTxXOFF(cp);
	}

	return(ESUCCESS);
}

/*-------------------------------------------------------------------------
 RCKTClose -  Close down a port and disable it from signaling that it has
  work to do.
|-------------------------------------------------------------------------*/
int RCKTClose(int PortNo)
{
	CHANNEL_T *cp;
	RCKTPort_t *Port;
	int pn;

	Port= &RcktPort[PortNo];

	if (!Port->Acquired)
		return(EFAILURE);

	cp= Port->cp;

	sDisRxFIFO(cp);
	sDisTransmit(cp);

	sDisInterrupts(cp, (MCINT_EN|RXINT_EN|SRCINT_EN|CHANINT_EN));

	sDisCTSFlowCtl(cp);
	sDisTxSoftFlowCtl(cp);
	sClrTxXOFF(cp);

	free(Port->cp);
	free(Port->rxq);
	free(Port->txq);

	Port->Acquired= 0;

	// build a compact, ordered list of open ports

	_disable();	// don't let the ISR do anything during this adventure
	OpenPortCount= 0;
	for (pn=0; pn<MAX_PORTS_PER_BOARD; pn++)
	{	if (RcktPort[pn].Acquired)
		{	OpenPortList[OpenPortCount]= pn;
			OpenPortCount++;
		}
	}
	_enable();
		
	return(ESUCCESS);
}

/*-------------------------------------------------------------------------
 RCKTTxFlush - Flush the transmit queue (both local and on the board)
|-------------------------------------------------------------------------*/
int RCKTTxFlush(int PortNo)
{
	RCKTPort_t *Port;
	CHANNEL_T *cp;
	unsigned char junk;

	Port= &RcktPort[PortNo];

	if (!Port->Acquired)
		return(EFAILURE);

	cp= Port->cp;

	sFlushTxFIFO(cp);
	sFlushTxPriorityBuf(cp, &junk);
	Port->txqCount= 0;
	Port->txqPut= 0;
	Port->txqGet= 0;

	return(ESUCCESS);
}

/*-------------------------------------------------------------------------
 RCKTRxFlush -
|-------------------------------------------------------------------------*/
int RCKTRxFlush(int PortNo)
{
	RCKTPort_t *Port;
	CHANNEL_T *cp;

	Port= &RcktPort[PortNo];

	if (!Port->Acquired)
		return(EFAILURE);

	cp= Port->cp;

	sFlushRxFIFO(cp);
	Port->rxqCount= 0;
	Port->rxqPut= 0;
	Port->rxqGet= 0;

	if (Port->rxFlowState & 1)
	{	sWriteTxPrioByte(cp, Port->receiveXON);
		Port->rxFlowState &= ~1;
	}

	if (Port->rxFlowState & 2)
	{	sSetRTS(cp);
		Port->rxFlowState &= ~2;
	}

	return(ESUCCESS);
}

/*-------------------------------------------------------------------------
 RCKTModemControl - Change the state of one of the modem control outputs
|-------------------------------------------------------------------------*/
int RCKTModemControl(int PortNo, int Signal, int State)
{
	CHANNEL_T *cp;
	RCKTPort_t *Port;

	Port= &RcktPort[PortNo];

	if (!Port->Acquired)
		return(EFAILURE);

	cp= Port->cp;

	if (State)
	{	if (Signal & MDMCTL_DTR)
			sSetDTR(cp);
		if (Signal & MDMCTL_RTS)
		{	if (Port->FlowCtrlMode & 2)
			{	if (Port->rxqCount < Port->rxqFlowOff)
					sSetRTS(cp);
			}
			else
				sSetRTS(cp);
		}
	}
	else
	{	if (Signal & MDMCTL_DTR)
			sClrDTR(cp);
		if (Signal & MDMCTL_RTS)
			sClrRTS(cp);
	}

	return(ESUCCESS);
}

/*-------------------------------------------------------------------------
 RCKTModemStatus - Determine the state of the modem control inputs
|-------------------------------------------------------------------------*/
int RCKTModemStatus(int PortNo)
{
	CHANNEL_T *cp;
	RCKTPort_t *Port;
	unsigned char ChanStatusLo;
	int modemStatus;
	
	Port= &RcktPort[PortNo];

	if (!Port->Acquired)
		return(EFAILURE);

	cp= Port->cp;

	ChanStatusLo= sGetChanStatusLo(cp);

	modemStatus =
			(ChanStatusLo & CTS_ACT) ? MDMSTA_CTS : 0;
	modemStatus |=
			(ChanStatusLo & DSR_ACT) ? MDMSTA_DSR : 0;
	modemStatus |=
			(ChanStatusLo &  CD_ACT) ? MDMSTA_DCD : 0;

	return(modemStatus);
}

/*-------------------------------------------------------------------------
 RCKTReadData - Read a block of data from a port's receive queue
|-------------------------------------------------------------------------*/
int RCKTReadData(int PortNo, char *Buffer, int Length, int *NumBytesRead)
{
	RCKTPort_t	*Port;
	int	rxqCount, rxqGet, rxqSize;
	int	TotalMove, Move1, Move2;
	CHANNEL_T *cp;

	Port= &RcktPort[PortNo];

	if (!Port->Acquired)
		return(EFAILURE);

	TotalMove= Length;

	rxqCount= Port->rxqCount;
	rxqGet= Port->rxqGet;
	rxqSize= Port->rxqSize;

	if (TotalMove > rxqCount)
		TotalMove= rxqCount;

	if (TotalMove)
	{
		Move2= TotalMove - (rxqSize - rxqGet);
		if (Move2 < 0)
			Move2= 0;
		Move1= TotalMove - Move2;

		memmove(Buffer, Port->rxq+rxqGet, Move1);
		if (Move2)
			memmove(Buffer+Move1, Port->rxq, Move2);

		Port->rxqGet += TotalMove;
		if (Port->rxqGet >=  Port->rxqSize)
			Port->rxqGet -= Port->rxqSize;

		Port->rxqCount -= TotalMove;
	}
	
//// handle receive flow control.  Flow on if we have enough room in
//// the receive queue.

	if (Port->rxFlowState && (Port->rxqCount <= Port->rxqFlowOn))
	{
		cp= Port->cp;

//// SW to be flowed on ?
		if (Port->rxFlowState & 1)
		{	sWriteTxPrioByte(cp, Port->receiveXON);
			Port->rxFlowState &= ~1;
		}

//// HW to be flowed on ?
		if (Port->rxFlowState & 2)
		{	sSetRTS(cp);
			Port->rxFlowState &= ~2;
		}
	}

	*NumBytesRead= TotalMove;

	return (ESUCCESS);
}

/*-------------------------------------------------------------------------
 RCKTWriteData - Write a block of data to a port's transmit queue
|-------------------------------------------------------------------------*/
int RCKTWriteData(int PortNo, char *Buffer, int Length, int *numBytesWritten)
{
	RCKTPort_t	*Port;
	int	txqFree, txqPut, txqSize, txqCount;
	int	TotalMove, Move1, Move2;

	Port= &RcktPort[PortNo];

	if (!Port->Acquired)
		return(EFAILURE);

	TotalMove= Length;

	txqPut= Port->txqPut;
	txqSize= Port->txqSize;
	txqCount= Port->txqCount;
	txqFree= txqSize - txqCount;

	if (TotalMove > txqFree)
		TotalMove= 0;

	if (TotalMove)
	{
		Move2= TotalMove - (txqSize - txqPut);
		if (Move2 < 0)
			Move2= 0;
		Move1= TotalMove - Move2;
		memmove(Port->txq + txqPut, Buffer, Move1);
		if (Move2)
			memmove(Port->txq, Buffer + Move1, Move2);

		Port->txqPut += TotalMove;
		if (Port->txqPut >= txqSize)
			Port->txqPut -= txqSize;

		Port->txqCount += TotalMove;
	}
	
	*numBytesWritten = TotalMove;

	return(ESUCCESS);
}

/*-------------------------------------------------------------------------
 RCKTRxPoll -
|-------------------------------------------------------------------------*/
void RCKTRxPoll(RCKTPort_t *Port)
{
	CHANNEL_T *cp;
	unsigned int ChanStatus, CharNStat;
	unsigned char RrxChar;
	int ToRecv, Recv1, Recv2, wRecv;
	int rxqFree, rxqPut;
   WORD *wordptr;

	cp= Port->cp;

//// How many are waiting at the door ?

	ToRecv= sGetRxCnt(cp);

//// read the interrupt ID register to clear the interrupt flag
	sGetChanIntID(cp);

//// read the status of this channel to determine if there are any
//// error'd characters in the FIFO

	ChanStatus= sGetChanStatus(cp);

	if (ToRecv)
	{

//// if status indicates there are errored characters in the FIFO, then
//// enter status mode (a word in FIFO holds character and status).

		if (ChanStatus & (RXFOVERFL | RXBREAK | RXFRAME | RXPARITY))
		{
			if (!(ChanStatus & STATMODE))
			{	ChanStatus |= STATMODE;
				sEnRxStatusMode(cp);
			}
		}
	
//// determine how many we can actually read in.  If we can't read any in
//// then we have a software overrun condition.

		rxqFree= Port->rxqSize - Port->rxqCount;
		if (ToRecv > rxqFree)
		{	if (rxqFree == 0)
			{	Port->overrunSoftware++;
				goto RxPoll1;
			}
			ToRecv= rxqFree;
		}

//// if we previously entered status mode, then read down the FIFO one
//// word at a time, pulling apart the character and the status.  Update
//// error counters depending on status

		if (ChanStatus & STATMODE)	// slow read because of error	
		{	
			while (ToRecv)
			{	CharNStat= sInW(sGetTxRxDataIO(cp));	

				if (CharNStat & STMFRAMEH)
					Port->framingErrors++;

				if (CharNStat & STMRCVROVRH)
					Port->overrunHardware++;

				if (CharNStat & STMPARITYH)
					Port->parityErrors++;

				RrxChar= CharNStat & 0xff;	// low byte is char
				Port->rxq[Port->rxqPut]= RrxChar;

//// bump the queue count and the Put pointer... handle wrap if neccessary

				Port->rxqPut++;
				if (Port->rxqPut >= Port->rxqSize)
					Port->rxqPut -= Port->rxqSize;
				Port->rxqCount++;
				ToRecv--;
			}

//// after we've emptied the FIFO in status mode, turn status mode back off

			if (sGetRxCnt(cp) == 0)
				sDisRxStatusMode(cp);
		}

//// we aren't in status mode, so read down the FIFO two characters at
//// time by doing repeated word IO transfer.

		else
		{
			rxqPut= Port->rxqPut;

//// figure out the two chunks (before and after wrap) that we will need
//// to move

			Recv2= ToRecv - (Port->rxqSize - rxqPut);
			if (Recv2 < 0)
				Recv2= 0;
			Recv1= ToRecv - Recv2;

//// move the first chunk... firstly by words and then the straggling odd byte

			wRecv= Recv1 >> 1;
			if (wRecv)
           {
              wordptr = (WORD *)(Port->rxq+rxqPut); // need a word ptr for this macro
				sInStrW(sGetTxRxDataIO(cp), wordptr, wRecv);
           }
			if (Recv1 & 1)
				Port->rxq[rxqPut+Recv1-1]= sInB(sGetTxRxDataIO(cp));

//// and then the second chunk

			wRecv= Recv2 >> 1;
			if (wRecv)
           {
              wordptr = (WORD *)(Port->rxq); // need a word ptr for this macro
				sInStrW(sGetTxRxDataIO(cp), wordptr, wRecv);
           }
			if (Recv2 & 1)
				Port->rxq[Recv2-1]= sInB(sGetTxRxDataIO(cp));

//// update Put pointer and queue count

			Port->rxqPut += ToRecv;
			if (Port->rxqPut >= Port->rxqSize)
				Port->rxqPut -= Port->rxqSize;
			Port->rxqCount += ToRecv;
		}

//// handle input flow control...  rxq getting full ?

RxPoll1:
		if (Port->rxqCount >= Port->rxqFlowOff)
		{
//// we need to flow off...

//// is SW flow enabled and not already flowed off ?

			if ((Port->FlowCtrlMode & 1) && !(Port->rxFlowState & 1))
			{

//// send the XOFF character out through the priorty queue

				sWriteTxPrioByte(cp, Port->receiveXOFF);
				Port->rxFlowState |= 1;
			}

//// is HW flow enabled and not already flowed off ?

			if ((Port->FlowCtrlMode & 2) && !(Port->rxFlowState & 2))
			{
//// drop CTS
				sClrRTS(cp);
				Port->rxFlowState |= 2;
			}
		}
	}
}

/*-------------------------------------------------------------------------
 RCKTTxPoll - Handle transmitter work during the poll interrupt
|-------------------------------------------------------------------------*/
void RCKTTxPoll(RCKTPort_t *Port)
{
	CHANNEL_T *cp;
	int txqGet, RtxqFree, ToSend, wSend, Send1, Send2;
   WORD *wordptr;

	cp= Port->cp;

//// How many would we really like to send ?

	ToSend= Port->txqCount;

//// How many can we actually send ?

	RtxqFree= TXFIFO_SIZE - (int)sGetTxCnt(cp);
	if (ToSend > RtxqFree)
		ToSend= RtxqFree;

//// if we can't send any, then the pipe is full (possibly flowed off)

	if (ToSend <= 0)
		return;

//// the block to send might actually be two smaller blocks. One before
//// and one after the end of the queue wraps.
//// determine Send1 (before the wrap) and Send2 (after the wrap) counts.

	txqGet= Port->txqGet;
	Send2= ToSend - (Port->txqSize - txqGet);
	if (Send2 < 0)
		Send2= 0;
	Send1= ToSend - Send2;

//// send the first chunk words at a time, and then any odd byte that might
//// be left

	wSend= Send1 >> 1;
	if (wSend)
   {
      wordptr = (WORD *)(Port->txq+txqGet); // need a word ptr for this macro
		sOutStrW(sGetTxRxDataIO(cp), wordptr, wSend); 
   }
	if (Send1 & 1)
		sOutB(sGetTxRxDataIO(cp), (unsigned char)*(Port->txq+txqGet+Send1-1));

//// and then send the second chunk as words and any remaining odd byte

	wSend= Send2 >> 1;
	if (wSend)
   {
      wordptr = (WORD *)(Port->txq);  // need a word ptr for this macro
		sOutStrW(sGetTxRxDataIO(cp), wordptr, wSend);
   }
	if (Send2 & 1)
		sOutB(sGetTxRxDataIO(cp), (unsigned char)*(Port->txq+Send2-1));

//// update the queue Get pointer and the total count remaining in the queue

	Port->txqGet += ToSend;
	if (Port->txqGet >= Port->txqSize)
		Port->txqGet -= Port->txqSize;
	Port->txqCount -= ToSend;
}

/*-------------------------------------------------------------------------
 RCKTPollISR - Handle the poll interrupt. This gets called each time the board
 makes a tick.  However, this could be called by an OS
 polling mechanism and might not be a real ISR in that case.
|-------------------------------------------------------------------------*/
void interrupt RCKTPollISR()
{
	CONTROLLER_T *ctlp;
	RCKTPort_t *Port;
	int aiop, ch, pn;
	unsigned char CtlServiceMask, AiopServiceMask;

//// we're dealing with the 0th controller

	ctlp= &sController[0];

//// fetch the status of interrupts on the board. This will tell us which
//// AIOPs have receive work to do.

	CtlServiceMask= sGetControllerIntStatus(ctlp);
	for (aiop=0; CtlServiceMask != 0; aiop++)
	{

//// if a particular AIOP needs service, then check which channels within
//// that AIOP have receive work to do.

		if (CtlServiceMask & 1)
		{	AiopServiceMask= sGetAiopIntStatus(ctlp, aiop);
			for (ch=0; AiopServiceMask != 0; ch++)
			{

//// if a particular channel needs service, then convert it's AIOP
//// and channel numbers to a port index for our structure and then
//// go do the RxPoll work for that port.

				if (AiopServiceMask & 1)
				{	pn= sGetDevMap(0,aiop,ch);
					RCKTRxPoll(&RcktPort[pn]);
				}
				AiopServiceMask >>= 1;
			}
		}
		CtlServiceMask >>= 1;
	}

//// run through the list of open ports and handle any transmit work
//// that might be pending

	for (pn=0; pn<OpenPortCount; pn++)
	{
		Port= &RcktPort[OpenPortList[pn]];

		if (Port->txqCount)
			RCKTTxPoll(Port);
	}

//// issue EOI's to anyone and everyone that needs them

	if (IrqNum > 7)			// slave 8259 ??
		outp(0xa0,0x20);
	outp(0x20,0x20);		// master 8259

	sControllerEOI(ctlp);		// EOI to the board
}

/*-------------------------------------------------------------------------
 htoi - convert a hex/ascii string to an unsigned int
|-------------------------------------------------------------------------*/
unsigned htoi(char *s)
{
	register unsigned i= 0;
	register int c;

	while ( (c= *s) != 0)
	{	c -= '0';
		if (c > 0x30)
			c -= 0x20;
		if (c > 0x11)
			c -= 7;
		if (c < 0 || c > 15)
			break;
		i <<= 4;
		i |= c;
		s++;
	}
	return(i);
}
