/******************************************************************************
File:      TPUT.C

Project:   RocketPort DOS API

Purpose:   Throughput test

Author:    CH

Date:      March 11, 1994

Operation: Install a loopback plugs on all ports to be tested.  At the DOS
           command line type:
              DBUF <First device> <Last device> <Display interval>
           First device is the first device (port) number to test.
           Last device is the last device number to test.
           Display interval is the interval, in seconds, between screen output
           of test results.  Larger display intervals give greater accuracy,
           a 10 second interval is adequate.

           At each display interval the test outputs one line of information
           for that interval as follows:
              Total Tx: WWWW  Total Rx: XXXX  Port Tx: YYYY  Port Rx: ZZZZ
           where:
              WWWW = The total number of characters transmitted per second by
                     the entire controller.  
              XXXX = The total number of characters received per second by
                     the entire controller.  
              YYYY = The average number of characters transmitted per second by
                     each port on the controller.  This is WWWW divided by the
                     number of devices under test.  
              ZZZZ = The average number of characters received per second by
                     each port on the controller.  This is XXXX divided by the
                     number of devices under test.  

Comments:  Time intervals are measured by the controller's periodic interrupt,
           which runs at about 281.1 Hz.  The fractional part of this interval
           is not taken into account by this test, so actual throughput rates
           will be a few bps higher than those displayed.  This error is too
           small to be significant.  
*******************************************************************************/


#include <stdio.h>
#include <dos.h>
#include <process.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include "api.h"

#define OUTBUF_SIZE 256
#define INBUF_SIZE  1024
#define BAUD_RATE   COM_BAUD_230400

int FirstDev, LastDev;
int PeriodicEvents, DisplayTrig;
long TotalTx, TotalRx;
int DispInterval;                      /* Interval between display of data */
unsigned char OutBuf[OUTBUF_SIZE + 1];
unsigned char InBuf[INBUF_SIZE];
unsigned char Str1[] = "This is a test of the Comtrol RocketPort total throughput, this is only a test!!";

void TestThroughput(void);
void EvPeriodic(void);

/******************************************************************************
Function: main
Purpose:  Initialization, start throughput test. 
*/
main(int argc,char *argv[])
{
   unsigned int InitRet;
   int i,j;

   /* Get command line parameters */
   if(argc != 4)
   {
      printf("Usage: tput <First device> <Last device> <Display Interval>\n");
      exit(1);
   }
   FirstDev = atoi(argv[1]);
   LastDev = atoi(argv[2]);
   DispInterval = atoi(argv[3]);
   if(DispInterval < 1)                /* avoid divide by zero error */
      DispInterval = 1;
   printf("First Device = %d, Last Device = %d, Display Interval = %d\n",
           FirstDev,LastDev,DispInterval);

   /* Configure API */
   aaInstallCtrlCHandler();
   if((InitRet = aaInit()) != NO_ERR)
   {
      printf("Init fail: %x\n",InitRet);
      exit(1);
   }

   /* Initialize output buffer */
   for(i = 0;i < OUTBUF_SIZE;i += 80)
   {
      j = 80;
      if(i + 80 > OUTBUF_SIZE)      /* this copy reaches end of buffer */
         j = OUTBUF_SIZE - i;
      strncpy((char *)&OutBuf[i],(char *)Str1,j);
   }

   /* Do the throughput test */
   TestThroughput();
   aaExit();
   return(0);
}

/******************************************************************************
Function: TestThroughput
Purpose:  Throughput test.
*/
void TestThroughput(void)
{
   int Dev;
   long Divisor;
   int Err;
   int AllowPrint = FALSE;

   /* Initialization */
   Divisor = (LastDev - FirstDev + 1) * DispInterval;
   DisplayTrig = (int)DispInterval * 281;
   PeriodicEvents = 0;
   TotalTx = 0L;
   TotalRx = 0L;
   aaInstallPeriodicEvent(EvPeriodic);
   aaEnPeriodicEvent(TRUE);

   /* Open all devices */
   for(Dev = FirstDev;Dev <= LastDev;Dev++)
   {
      if((Err = aaOpen(Dev,COM_TX | COM_RX,BAUD_RATE,COM_PAR_NONE,COM_DATABIT_8,
                COM_STOPBIT_1,0,COM_DEN_NONE,COM_MDM_RTS | COM_MDM_DTR)) != 0)
      {
         printf("Open Failure - Device %d, Error %d\n",Dev,Err);
         aaExit();
         exit(0);
      }
   }

   /* Test loop */
   while(!kbhit())
   {
      for(Dev = FirstDev;Dev <= LastDev;Dev++)  /* write and read max possible */
      {
         TotalTx += aaWrite(Dev,250,OutBuf);
         TotalRx += aaRead(Dev,1024,InBuf);
      }

      if(PeriodicEvents >= DisplayTrig)         /* display interval elapsed */
      {
         asm cli;

         /* First print has invalid info because the FIFOs are just getting
            filled up, block it */
         if(AllowPrint)
         {
            printf("Total Tx: %ld,  Total Rx: %ld,  Port Tx: %ld,  Port Rx: %ld\n",
                TotalTx/DispInterval,TotalRx/DispInterval,TotalTx/Divisor,TotalRx/Divisor);
         }
         else
            AllowPrint = TRUE;
         PeriodicEvents = 0;
         TotalTx = 0;
         TotalRx = 0;
         asm sti;
      }
   }
   getch();
}

/*****************************************************************************
Function: EvPeriodic
Purpose:  Periodic event function
*/
void EvPeriodic(void)
{
   PeriodicEvents++;
}

